null2na <- function(x) {
	if(is.null(x) || length(x)==0)
		x <- as.character(NA)
	else
		x <- unique(unlist(x))
	x
}

parseExperiment <- function(root, namespaces) {
	subDoc <- xmlDoc(root)
	interactionType <- unlist(xpathApply(doc=subDoc,
					path="/ns:experimentDescription/ns:interactionDetectionMethod/ns:names/ns:shortLabel",
					fun=xmlValue,
					namespaces=namespaces ))
	# extract the "id" attribute value
	expPubMed <- unlist(xpathApply(doc=subDoc,
				path="/ns:experimentDescription/ns:bibref/ns:xref/ns:primaryRef[@db='pubmed']",
				fun=xmlGetAttr,
				name="id",
				namespaces=namespaces))
	expIntAct <- unlist(xpathApply(doc=subDoc,
				path="/ns:experimentDescription/ns:xref/ns:primaryRef[@db='intact']|/ns:experimentList/ns:experimentDescription/ns:xref/ns:secondaryRef[@db='intact']",
				fun=xmlGetAttr,
				name="id",
				namespaces=namespaces))
	free(subDoc)

	list(	interactionType=null2na(interactionType),
		expPubMed=null2na(expPubMed),
		expIntAct=null2na(expIntAct)
	)
}

parseInteractor <- function(root, namespaces) {
	subDoc <- xmlDoc(root) 
	# IntAct ID of the interactor
	intactIds <- unlist(xpathApply(doc=subDoc,
				path="/ns:interactor/ns:xref/ns:secondaryRef[@db='intact' and not(@refType='isoform-parent')][1]|/ns:interactor/ns:xref/ns:primaryRef[@db='intact']",
				fun=xmlGetAttr,
				name="id",
				namespaces=namespaces))
	# Uniprot ID 
	uniprotIds <- unlist(xpathApply(doc=subDoc,
				path="/ns:interactor/ns:xref/ns:primaryRef[@db='uniprotkb' and @refType='identity']",
				fun=xmlGetAttr,
				name="id",
				namespaces=namespaces))
	# gene symbol
	geneNames <- unlist(xpathApply(doc=subDoc, 
				path="/ns:interactor/ns:names/ns:alias[@type='gene name'][1]",
				fun=xmlValue,
				namespaces=namespaces))
	# full name
	fullNames <- unlist(xpathApply(doc=subDoc, 
				path="/ns:interactor/ns:names/ns:fullName",
				fun=xmlValue,
				namespaces=namespaces))
	# locus name
	locusNames <- unlist(xpathApply(doc=subDoc, 
				path="/ns:interactor/ns:names/ns:alias[@type='locus name'][1]",
				fun=xmlValue,
				namespaces=namespaces))
	# orf name
	orfNames <- unlist(xpathApply(doc=subDoc, 
				path="/ns:interactor/ns:names/ns:alias[@type='orf name'][1]",
				fun=xmlValue,
				namespaces=namespaces))
	# organism name	
	organismNames <- unlist(xpathApply(doc=subDoc,
				path="/ns:interactor/ns:organism/ns:names/ns:fullName",
				fun=xmlValue,
				namespaces=namespaces))
	# taxonomy ID
	taxIds <- unlist(xpathApply(doc=subDoc,
				path="/ns:interactor/ns:organism",
				fun=xmlGetAttr,
				name="ncbiTaxId",
				namespaces=namespaces))
	free(subDoc)
	list(	intactId=intactIds, 
		uniprotId=uniprotIds, 
		geneName=geneNames,
		fullName=fullNames,
		locusName=locusNames,
		orfName=orfNames,
		organismName=organismNames,
		taxId=taxIds )
}

psi25interaction <- function(intactFile) {
    intactDoc <- xmlTreeParse(intactFile, useInternalNodes=TRUE)
    intactNS <- xmlNamespaceDefinitions(xmlRoot(intactDoc))
    namespaces <- c(ns=intactNS[[1]]$uri)

    entry <- getNodeSet(intactDoc, "/ns:entrySet/ns:entry", namespaces)
    entryCount <- length(entry)

    entryList <- lapply( seq(entryCount), function(i) {
	basePath <- paste("/ns:entrySet/ns:entry[", i, "]", sep="", collapse="")
	thisEntry <- new("interactionEntry")

	##################
	#    experiment  #
	##################
	experimentPath <- paste(basePath,
		"/ns:experimentList/ns:experimentDescription", sep="", collapse="")
	experimentNodes <- getNodeSet(intactDoc, experimentPath, namespaces)
	experimentEnv <- new.env(parent=emptyenv(), hash=TRUE)
	lapply(experimentNodes, function(thisExperiment) {
		experimentId <- xmlGetAttr(thisExperiment, name="id")
		experimentData <- parseExperiment(thisExperiment, namespaces)
		assign(experimentId, experimentData, envir=experimentEnv)
		""
	})

	################
	# releaseDate  #
	################			
	releaseDatePath <- paste(basePath, "/ns:source", sep="", collapse="")
	releaseDate <- xpathApply(doc=intactDoc,
				path=releaseDatePath,
				fun=xmlGetAttr,
				name="releaseDate",
				namespaces=namespaces)
	thisEntry@releaseDate <- null2na(releaseDate)

	##############
	# interactor #
	##############
	# ID of the interactor, only unique within the XML doc
	interactorPath <- paste(basePath,
		"/ns:interactorList/ns:interactor", sep="", collapse="")
	interactorNodes <- getNodeSet(intactDoc, interactorPath, namespaces)
	interactorIds <- sapply(interactorNodes, xmlGetAttr, name="id")
	interactorCount <- length(interactorNodes)
	interactors <- matrix(character(0), interactorCount, 8)
	colnames(interactors) <- c("intactId", "uniprotId", "geneName",
                                   "fullName", "locusName", "orfName",
                                   "organismName", "taxId")
	for (i in seq(interactorCount)) {
		theRes <- parseInteractor(interactorNodes[[i]], namespaces)
		# remove "NULL" and "list()"
		theRes <- unlist(theRes, recursive=FALSE)
		interactors[i, names(theRes)] <- theRes 
	}
		
	###################
	# organismName	  #
	###################
	organismName <- unique(unlist(interactors[,"organismName"]))
	thisEntry@organismName <- organismName

	#########
	# taxId #
	#########
	taxId <- unique(unlist(interactors[,"taxId"]))
	thisEntry@taxId <- taxId

	#################
	# interactions  #
	#################
	# ID of the interacton, only unique within the XML doc
	interactionPath <- paste(basePath,
		"/ns:interactionList/ns:interaction", sep="", collapse="")
	interactionNodes <- getNodeSet(intactDoc, interactionPath, namespaces=namespaces)
	interactions <- lapply(interactionNodes, function(theNode) {
				subDoc <- xmlDoc(theNode)
				# interaction ID
				intactId <- xpathApply(doc=subDoc, 
					path="/ns:interaction/ns:xref/ns:primaryRef[@db='intact']",
					fun=xmlGetAttr,
					name="id",
					namespaces=namespaces)[[1]]
				# experiment reference
				expRef <- unlist(xpathApply(doc=subDoc,
					path="/ns:interaction/ns:experimentList/ns:experimentRef[1]",
					fun=xmlValue,
					namespaces=namespaces))
				if ((!is.null(expRef)) && exists(expRef, envir=experimentEnv)) {
					expData <- get(expRef, envir=experimentEnv)
					interactionType <- expData$interactionType
					expIntAct <- expData$expIntAct
					expPubMed <- expData$expPubMed
				} else {
					interactionType <- as.character(NA)
					expIntAct <- as.character(NA)
					expPubMed <- as.character(NA)
				}
				# confidence value
				confidenceValue <- unlist(xpathApply(doc=subDoc, 
					path="/ns:interaction/ns:confidenceList/ns:confidence/ns:value",
					fun=xmlValue,
					namespaces=namespaces))
				if (is.null(confidenceValue))
					confidenceValue <- as.character(NA)
				# get the internal ID of the interactors
				rolePath <- "/ns:interaction/ns:participantList/ns:participant[ns:experimentalRoleList/ns:experimentalRole/ns:names/ns:fullName='bait']/ns:interactorRef" 
				roleIds <- xpathApply(doc=subDoc,
							path=rolePath,
							fun=xmlValue,
							namespaces=namespaces)
				baitRefs <- unlist(roleIds)
				rolePath <- "/ns:interaction/ns:participantList/ns:participant[ns:experimentalRoleList/ns:experimentalRole/ns:names/ns:fullName='prey']/ns:interactorRef" 
				roleIds <- xpathApply(doc=subDoc,
							path=rolePath,
							fun=xmlValue,
							namespaces=namespaces)
				preyRefs <- unlist(roleIds)
				rolePath <- "/ns:interaction/ns:participantList/ns:participant[ns:experimentalRoleList/ns:experimentalRole/ns:names/ns:fullName='inhibitor']/ns:interactorRef" 
				roleIds <- xpathApply(doc=subDoc,
							path=rolePath,
							fun=xmlValue,
							namespaces=namespaces)
				inhibitorRefs <- unlist(roleIds)
				rolePath <- "/ns:interaction/ns:participantList/ns:participant[ns:experimentalRoleList/ns:experimentalRole/ns:names/ns:fullName='neutral component']/ns:interactorRef" 
				roleIds <- xpathApply(doc=subDoc,
							path=rolePath,
							fun=xmlValue,
							namespaces=namespaces)
				neutralComponentRefs <- unlist(roleIds)
				
				free(subDoc)
				# map to the intact ID of interactors
				mapColName <- "intactId"
				baitIndex <- which(interactorIds %in% baitRefs)
				if (length(baitIndex)==0)
					baitUniprot <- as.character(NA)
				else
					baitUniprot <- unlist(interactors[baitIndex, mapColName])

				preyIndex <- which(interactorIds %in% preyRefs)
				if (length(preyIndex)==0)
					preyUniprot <- as.character(NA)
				else
					preyUniprot <- unlist(interactors[preyIndex, mapColName])
				inhibitorIndex <- which(interactorIds %in%  inhibitorRefs)
				if (length(inhibitorIndex)==0)
					inhibitorUniprot <- as.character(NA)
				else
					inhibitorUniprot <- unlist(interactors[inhibitorIndex, mapColName])

				neutralComponentIndex <- which(interactorIds %in% neutralComponentRefs)
				if (length(neutralComponentIndex)==0)
					neutralComponentUniprot <- as.character(NA)
				else
					neutralComponentUniprot <- unlist(interactors[neutralComponentIndex, mapColName])

				new("intactInteraction", 
					intact=intactId, 
					interactionType=interactionType,
					expPubMed=expPubMed,
					expIntAct=expIntAct,
					confidenceValue=confidenceValue,
					bait=baitUniprot,
					prey=preyUniprot,
					inhibitor=inhibitorUniprot,
					neutralComponent=neutralComponentUniprot)

	})
	thisEntry@interactions <- interactions
	rownames(interactors) <- interactors[, "intactId"]
	interactors <- interactors[,-1]
	thisEntry@interactors <- interactors
	thisEntry
    })
    free(intactDoc)
    entryList <- entryList[[1]]
    entryList
}

parseComplex <- function(root, namespaces) {
	subDoc <- xmlDoc(root)
	# intact ID
	intactId <- as.character(xpathApply(doc=subDoc,
				path="/ns:interaction/ns:xref/ns:primaryRef[@db='intact']",
				fun=xmlGetAttr,
				name="id",
				namespaces=namespaces))
	# short label
	shortLabel <- as.character(xpathApply(doc=subDoc,
				path="/ns:interaction/ns:names/ns:shortLabel",
				fun=xmlValue,
				namespaces=namespaces))
	# full name
	fullName <-as.character(xpathApply(doc=subDoc,
                                path="/ns:interaction/ns:names/ns:fullName",
                                fun=xmlValue,
                                namespaces=namespaces))
 
	# interactors
	interactorIds <-as.character(xpathApply(doc=subDoc,
                                path="/ns:interaction/ns:participantList/ns:participant/ns:interactorRef",
                                fun=xmlValue,
                                namespaces=namespaces))

	# attributes
	attributeNodes <- getNodeSet(subDoc, 
				"/ns:interaction/ns:attributeList/ns:attribute[@name]",
				namespaces=namespaces)
	if (length(attributeNodes)>0) {
		attributeNames <- sapply(attributeNodes, xmlGetAttr, name="name")
		attributes <- sapply(attributeNodes, xmlValue)
		names(attributes) <- attributeNames
	} else {
		attributes <- character(0)
	}
	free(subDoc)
	
	list(intactId=intactId, shortLabel=shortLabel, fullName=fullName, interactorIds=interactorIds, attributes=attributes)
}

psi25complex <- function(intactFile) {
    intactDoc <- xmlTreeParse(intactFile, useInternalNodes=TRUE)
    intactNS <- xmlNamespaceDefinitions(xmlRoot(intactDoc))
    namespaces <- c(ns=intactNS[[1]]$uri)

    releaseDate <- xpathApply(doc=intactDoc, path="//ns:entry/ns:source", 
				fun=xmlGetAttr, name="releaseDate", 
				namespaces=namespaces)[[1]]
    	##############
	# interactor #
	##############
	interactorNodes <- getNodeSet(intactDoc, "//ns:interactorList/ns:interactor", namespaces)
	interactorIds <- sapply(interactorNodes, xmlGetAttr, name="id")
	interactorCount <- length(interactorNodes)
	interactors <- matrix(character(0), interactorCount, 8)
	colnames(interactors) <- c("intactId", "uniprotId", "geneName", "fullName", "locusName", "orfName", "organismName", "taxId")
	for (i in seq(interactorCount)) {
		theRes <- parseInteractor(interactorNodes[[i]], namespaces)
		# remove "NULL" and "list()"
		theRes <- unlist(theRes, recursive=FALSE)
		interactors[i, names(theRes)] <- theRes 
	}

    	##############
	# complex    #
	##############
	complexNodes <- getNodeSet(intactDoc, "//ns:interactionList/ns:interaction", namespaces)
	complexList <- lapply(complexNodes, function(thisNode) {
				thisComplex <- parseComplex(thisNode, namespaces)
				participantRef <- split(thisComplex$interactorIds, 
							as.factor(thisComplex$interactorIds))
				multiplicity <- sapply(participantRef, length)
				if ( !all(names(participantRef) %in% interactorIds) ) {
					msg <- paste("parse complex ", thisComplex$shortLabel, 
							"can't resolve all interactor references.")
					stop(msg)
				}	
				participantIndex <- interactorIds %in% names(participantRef)
				participants <- interactors[participantIndex, "intactId"]
				participants <- data.frame(intActId=participants, multiplicity=multiplicity)
				thisOrganism <- unique(as.character(interactors[participantIndex,"organismName"]))
				thisTaxid <- unique(as.character(interactors[participantIndex, "taxId"]))
				new( "intactComplex",
					intactId=thisComplex$intactId,
					shortLabel=thisComplex$shortLabel,
					fullName=thisComplex$fullName,
					organismName=thisOrganism,
					taxId=thisTaxid,
					members=participants,
					attributes=thisComplex$attributes
				)
			})
    free(intactDoc)
    rownames(interactors) <- interactors[, "intactId"]
    interactors <- interactors[,-1]
    new("complexEntry", interactors=interactors, complexes=complexList, releaseDate=releaseDate)
}

intactXML2Graph <- function(intactFiles, type="interaction", directed = TRUE){
  if(type == "interaction"){

    result <- lapply(intactFiles, psi25interaction)
    
    baitList <- lapply(result, function(x){
      baits <- sapply(x@interactions, function(y) y@bait)
    })
    
    preyList <- lapply(result, function(x){
      prey <- sapply(x@interactions, function(y) y@prey)
    })

    index <- sapply(baitList, function(x) class(x)) == sapply(preyList, function(x) class(x))

    for(i in 1:length(index)){

        if(!index[i]){
            newBait <- vector(length=length(unlist(preyList[[i]])))
            k <- 1
            for(j in 1:length(preyList[[i]])){
                newBait[k:(k+length(preyList[[i]][[j]])-1)] <- rep(baitList[[i]][j], length(preyList[[i]][[j]]))
                k <- k + length(preyList[[i]][[j]])
            }
            baitList[[i]] <- newBait
        }

    }

    
    b <- unlist(baitList)
    p <- unlist(preyList)
    
    bpList <- split(p,b)
    bpMat <- list2Matrix(bpList)
    
    bpG <- genBPGraph(bpMat, directed = directed)
    
    l <- lapply(result, function(x) return(x@interactors))
    
    mat <- do.call(rbind, l)
    rn <- lapply(l, rownames)
    uRN <- vector()
    for(i in 1:length(rn)) uRN <- union(uRN, rn[[i]])
    mat <- mat[uRN,,drop=FALSE]
    
    bpGraph <- as(bpG, "intactGraph")
    bpGraph@interactors <- mat
    
    
    
  }
  
  if (type == "complex"){
    result <- lapply(intactFiles, psi25complex)

    listOfListOfComps <- lapply(result, function(x){
      lapply(x@complexes, function(y){
        p <- as.vector(rep(y@members[,1], y@members[,2]))
        
        attr(p, "complexName") <- y@fullName
        attr(p, "intactId") <- y@intactId
        p
        })
      })
    he <- do.call(c, listOfListOfComps)
    nodes <- unique(unlist(he))
    hEdges <- lapply(he, function(x) {e <- Hyperedge(x, attr(x,"complexName"))})
    intactID <- sapply(he, function(x){return(attr(x, "intactId"))})
    names(hEdges) <- intactID
    #bpG<- new("Hypergraph", nodes=nodes, hyperedges=hEdges)

    l <- lapply(result, function(x) return(x@interactors))
    
    mat <- do.call(rbind, l)
    rn <- lapply(l, rownames)
    uRN <- vector()
    for(i in 1:length(rn)) uRN <- union(uRN, rn[[i]])
    mat <- mat[uRN,,drop=FALSE]

    bpGraph <- new("intactHyperGraph", interactors=mat, nodes=nodes,
                   hyperedges = hEdges)
  }

  return(bpGraph)

}


separateXMLDataByExpt <- function(intactFiles, type="direct", directed=TRUE){
  options(error=recover)
  ie <- lapply(intactFiles, psi25interaction)
  interactorM <- lapply(ie, function(x) x@interactors)

  translationM <- do.call(rbind, interactorM)
  intactID <- unique(rownames(translationM))
  translationM1 <- translationM[intactID,]
  
  intactL = lapply(ie, function(x){
    getbpPairsFrominteractionEntry2(x,type=type)})

  intact = do.call(rbind,intactL)
  colnames(intact) = c("bait","prey","expPubMed","expIntact")
  rownames(intact) = NULL
  intactBPPairs = split(data.frame(intact[,c("bait","prey")],
    stringsAsFactors=FALSE),intact[,"expIntact"])
  
  intactBPList <- lapply(intactBPPairs, function(x){
    split(x[,2],x[,1])})

  intactBPList <- intactBPList[listLen(intactBPList)>0]

  intactBPMat <- lapply(intactBPList, list2Matrix)
  intactBPMat <- lapply(intactBPMat, t)
  intactBPGraphs <- lapply(intactBPMat, function(x)
                           genBPGraph(x,
                                      directed=directed))

  intactGraphs <- lapply(intactBPGraphs, function(x) {
    y <- as(x, "intactGraph"); y@interactors <- translationM1;
  return(y)})

  return(intactGraphs)
  
}

getbpPairsFrominteractionEntry2 = function(interactionEntry, type = "direct")
{

  if(!(type %in% c("direct","indirect"))){
    stop("The argument type needs to be either direct or indirect")
  }
  
  if(type == "direct"){
    interactionTypeWanted = c("two hybrid","two hybrid array", 
                               "2h fragment pooling", "2 hybrid")}
  if(type == "indirect"){
    interactionTypeWanted = c("coip","anti tag coip","pull down","tap",
        "anti bait coip","affinity chrom","chromatography","ion exchange chrom")
  }
    interactions = interactionEntry@interactions
    bpPairs = lapply(interactions,function(x){
                if(any(x@interactionType %in% interactionTypeWanted)){
                      cbind(x@bait,x@prey,x@expPubMed,x@expIntAct);}
              })
    do.call(rbind,bpPairs)
}

genBPGraph <- function(bpMat, directed=TRUE, bp=TRUE){

  bpMat1 <- bpMat
  b <- rownames(bpMat)
  p <- colnames(bpMat)

  if(!bp){
    if(sum(b != p) != 0){
      stop("The rownames and the colnames must be identical.")
    }
  }

  else{
    baits <- union(rownames(bpMat), colnames(bpMat))
    preys <- baits

    bpMat1 <- matrix(0, length(baits), length(preys))
    dimnames(bpMat1) <- list(baits, preys)
    bpMat1[b,p] <- bpMat
    if(!directed) {
      bpMat1 <- bpMat1 + t(bpMat1)
      mode(bpMat1) <- "logical"
      mode(bpMat1) <- "numeric"
    }
  }

  if(directed){
    bpGraph <- as(bpMat1, "graphNEL")
  }

  else{
    bpGraph <- as(bpMat1, "graphNEL")
    bpGraph <- ugraph(removeSelfLoops(bpGraph))
  }
  
  bpGraph
  
}
